from .roa_checker import ROAChecker
from .roa import ROA
from .enums_and_dataclasses import ROARouted, ROAValidity, ROAOutcome

__all__ = ["ROAChecker", "ROARouted", "ROAValidity", "ROAOutcome", "ROA"]



from functools import cache
from ipaddress import IPv4Network, IPv6Network, ip_network

from .enums_and_dataclasses import ROAOutcome
from .roa import ROA
from .roa_trie import ROATrie
from .roa_tries import IPv4ROATrie, IPv6ROATrie


class ROAChecker:
    """Gets validity of prefix origin pairs against ROAs"""

    def __init__(self):
        """Initializes both ROA tries"""

        self.ipv4_trie = IPv4ROATrie()
        self.ipv6_trie = IPv6ROATrie()
        self.get_roa_outcome_w_prefix_str_cached.cache_clear()

    def insert(self, prefix: IPv4Network | IPv6Network, roa: ROA) -> None:
        """Inserts a prefix into the tries"""

        trie = self.ipv4_trie if prefix.version == 4 else self.ipv6_trie
        # mypy struggling with this
        return trie.insert(prefix, roa)  # type: ignore

    def get_relevant_roas(self, prefix: IPv4Network | IPv6Network) -> frozenset[ROA]:
        """Gets the ROA covering prefix-origin pair"""

        trie = self.ipv4_trie if prefix.version == 4 else self.ipv6_trie
        assert isinstance(trie, ROATrie)
        return trie.get_relevant_roas(prefix)

    def get_roa_outcome(
        self, prefix: IPv4Network | IPv6Network, origin: int
    ) -> ROAOutcome:
        """Gets the validity of a prefix origin pair"""

        trie = self.ipv4_trie if prefix.version == 4 else self.ipv6_trie
        assert isinstance(trie, ROATrie), "for mypy"
        return trie.get_roa_outcome(prefix, origin)

    # NOTE: since this is called so often, we leave it as cache
    # since it's significantly faster
    # It literally says cached in the name, this better not cause leaks
    @cache  # noqa: B019
    def get_roa_outcome_w_prefix_str_cached(
        self,
        prefix_str: str,
        origin: int,
    ) -> ROAOutcome:
        """Returns ROA outcome for a prefix (str) and origin, and caches the result"""

        return self.get_roa_outcome(ip_network(prefix_str), origin)

    def clear(self):
        self.__init__()  # type: ignore



from ipaddress import IPv4Network, IPv6Network

from lib_cidr_trie import CIDRNode

from .roa import ROA


class ROAsNode(CIDRNode):
    def __init__(self, *args, **kwargs):
        """Initializes the ROA node"""

        super().__init__(*args, **kwargs)
        self.roas: set[ROA] = set()

    # Mypy doesn't understand *args in super class
    def add_data(  # type: ignore
        self, prefix: IPv4Network | IPv6Network, roa: ROA
    ) -> None:
        """Adds ROA to the node for that prefix"""

        self.prefix: IPv4Network | IPv6Network = prefix
        self.roas.add(roa)



from ipaddress import ip_network

from roa_checker import ROA, ROARouted, ROAValidity


def test_roa_outcome_valid():
    roa = ROA(ip_network("1.2.0.0/16"), 1)
    validity = roa.get_validity(ip_network("1.2.0.0/16"), 1)
    assert validity == ROAValidity.VALID
    outcome = roa.get_outcome(ip_network("1.2.0.0/16"), 1)
    assert outcome.validity == ROAValidity.VALID
    assert outcome.routed_status == ROARouted.ROUTED
    assert roa.routed_status == ROARouted.ROUTED
    assert roa.is_routed is True
    assert roa.is_non_routed is False


def test_roa_outcome_unknown():
    roa = ROA(ip_network("1.3.0.0/16"), 1)
    validity = roa.get_validity(ip_network("1.2.0.0/16"), 2)
    assert validity == ROAValidity.UNKNOWN
    outcome = roa.get_outcome(ip_network("1.2.0.0/16"), 2)
    assert outcome.validity == ROAValidity.UNKNOWN
    assert outcome.routed_status == ROARouted.ROUTED
    assert roa.routed_status == ROARouted.ROUTED
    assert roa.is_routed is True
    assert roa.is_non_routed is False


def test_roa_validity_invalid_length():
    roa = ROA(ip_network("1.2.0.0/16"), 1)
    validity = roa.get_validity(ip_network("1.2.3.0/24"), 1)
    assert validity == ROAValidity.INVALID_LENGTH
    outcome = roa.get_outcome(ip_network("1.2.3.0/24"), 1)
    assert outcome.validity == ROAValidity.INVALID_LENGTH
    assert outcome.routed_status == ROARouted.ROUTED
    assert roa.routed_status == ROARouted.ROUTED
    assert roa.is_routed is True
    assert roa.is_non_routed is False


def test_roa_validity_invalid_origin():
    roa = ROA(ip_network("1.2.0.0/16"), 1)
    validity = roa.get_validity(ip_network("1.2.0.0/16"), 0)
    assert validity == ROAValidity.INVALID_ORIGIN
    outcome = roa.get_outcome(ip_network("1.2.0.0/16"), 0)
    assert outcome.validity == ROAValidity.INVALID_ORIGIN
    assert outcome.routed_status == ROARouted.ROUTED
    assert roa.routed_status == ROARouted.ROUTED
    assert roa.is_routed is True
    assert roa.is_non_routed is False


def test_roa_validity_invalid_length_and_origin():
    roa = ROA(ip_network("1.2.0.0/16"), 1)
    validity = roa.get_validity(ip_network("1.2.3.0/24"), 2)
    assert validity == ROAValidity.INVALID_LENGTH_AND_ORIGIN
    outcome = roa.get_outcome(ip_network("1.2.3.0/24"), 2)
    assert outcome.validity == ROAValidity.INVALID_LENGTH_AND_ORIGIN
    assert outcome.routed_status == ROARouted.ROUTED
    assert roa.routed_status == ROARouted.ROUTED
    assert roa.is_routed is True
    assert roa.is_non_routed is False


def test_roa_non_routed():
    roa = ROA(ip_network("1.2.0.0/16"), 0)
    outcome = roa.get_outcome(ip_network("1.2.3.0/24"), 2)
    assert outcome.validity == ROAValidity.INVALID_LENGTH_AND_ORIGIN
    assert outcome.routed_status == ROARouted.NON_ROUTED
    assert roa.is_routed is False
    assert roa.is_non_routed is True



from dataclasses import dataclass
from functools import cached_property
from ipaddress import IPv4Network, IPv6Network

from .enums_and_dataclasses import ROAOutcome, ROARouted, ROAValidity


@dataclass(frozen=True)
class ROA:
    prefix: IPv4Network | IPv6Network
    origin: int
    max_length: int = None  # type: ignore
    # RIPE, ARIN, etc. Used by ROACollector, don't remove
    ta: str | None = None

    def __post_init__(self) -> None:
        if self.max_length is None:
            object.__setattr__(  # type: ignore
                self, "max_length", self.prefix.prefixlen
            )

    @cached_property
    def routed_status(self) -> ROARouted:
        return ROARouted.ROUTED if self.is_routed else ROARouted.NON_ROUTED

    @cached_property
    def is_routed(self) -> bool:
        return self.origin != 0

    @cached_property
    def is_non_routed(self) -> bool:
        return not self.is_routed

    def covers_prefix(self, prefix: IPv4Network | IPv6Network) -> bool:
        """Returns True if the ROA covers the prefix"""

        # NOTE: subnet_of includes the original prefix (I checked lol)
        # mypy wants this to be the same type of prefix but this doesn't matter
        return prefix.subnet_of(self.prefix)  # type: ignore

    def get_validity(
        self, prefix: IPv4Network | IPv6Network, origin: int
    ) -> ROAValidity:
        """Returns validity of prefix origin pair"""

        if self.covers_prefix(prefix):
            if prefix.prefixlen > self.max_length and origin != self.origin:
                return ROAValidity.INVALID_LENGTH_AND_ORIGIN
            elif prefix.prefixlen > self.max_length and origin == self.origin:
                return ROAValidity.INVALID_LENGTH
            elif prefix.prefixlen <= self.max_length and origin != self.origin:
                return ROAValidity.INVALID_ORIGIN
            elif prefix.prefixlen <= self.max_length and origin == self.origin:
                return ROAValidity.VALID
            else:
                raise NotImplementedError("This should never happen")
        else:
            return ROAValidity.UNKNOWN

    def get_outcome(self, prefix: IPv4Network | IPv6Network, origin: int) -> ROAOutcome:
        """Returns outcome of prefix origin pair"""

        validity = self.get_validity(prefix, origin)
        return ROAOutcome(validity=validity, routed_status=self.routed_status)
